@@ -25,14 +25,16 @@
 ###############################################################################
 
 import binascii
+from binascii import a2b_hex, b2a_hex
 import struct
-from typing import Callable, Optional, Union
+from typing import Callable, Optional, Union, Dict, Any
 
 import txaio
 
 from autobahn import util
-from autobahn.wamp.interfaces import IEd25519Key, ISession
+from autobahn.wamp.interfaces import ISecurityModule, ICryptosignKey
 from autobahn.wamp.types import Challenge
+from autobahn.wamp.message import _URI_PAT_REALM_NAME_ETH
 
 __all__ = [
     'HAS_CRYPTOSIGN',
@@ -41,11 +43,12 @@ __all__ = [
 try:
     # try to import everything we need for WAMP-cryptosign
     from nacl import encoding, signing, bindings
+    from nacl.signing import SignedMessage
 except ImportError:
     HAS_CRYPTOSIGN = False
 else:
     HAS_CRYPTOSIGN = True
-    __all__.append('SigningKey')
+    __all__.append('CryptosignKey')
 
 
 def _unpack(keydata):
@@ -294,24 +297,26 @@ def _qrcode_from_signify_ed25519_pubkey(pubkey_file, mode='text'):
     """
     assert(mode in ['text', 'svg'])
 
-    import pyqrcode
+    import qrcode
 
     with open(pubkey_file) as f:
         pubkey = f.read().splitlines()[1]
 
-        qr = pyqrcode.create(pubkey, error='L', mode='binary')
+        qr = qrcode.QRCode(box_size=3,
+                           error_correction=qrcode.ERROR_CORRECT_L)
+        qr.add_data(pubkey)
 
         if mode == 'text':
-            return qr.terminal()
-
-        elif mode == 'svg':
             import io
-            data_buffer = io.BytesIO()
 
-            qr.svg(data_buffer, omithw=True)
-
-            return data_buffer.getvalue()
+            with io.StringIO() as data_buffer:
+                qr.print_ascii(out=data_buffer, invert=True)
+                return data_buffer.getvalue()
+        elif mode == 'svg':
+            import qrcode.image.svg
 
+            image = qr.make_image(image_factory=qrcode.image.svg.SvgImage)
+            return image.to_string()
         else:
             raise Exception('logic error')
 
@@ -345,7 +350,7 @@ def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
     verify_key.verify(message, sig)
 
 
-# SigningKey from
+# CryptosignKey from
 #   - raw byte string or file with raw bytes
 #   - SSH private key string or key file
 #   - SSH agent proxy
@@ -356,7 +361,7 @@ def _verify_signify_ed25519_signature(pubkey_file, signature_file, message):
 
 if HAS_CRYPTOSIGN:
 
-    def format_challenge(challenge: Challenge, channel_id_raw: Optional[bytes], channel_id_type: Optional[str]) -> bytes:
+    def _format_challenge(challenge: Challenge, channel_id_raw: Optional[bytes], channel_id_type: Optional[str]) -> bytes:
         """
         Format the challenge based on provided parameters
 
@@ -400,7 +405,7 @@ if HAS_CRYPTOSIGN:
 
         return data
 
-    def sign_challenge(data: bytes, signer_func: Callable) -> bytes:
+    def _sign_challenge(data: bytes, signer_func: Callable) -> bytes:
         """
         Sign the provided data using the provided signer.
 
@@ -430,53 +435,64 @@ if HAS_CRYPTOSIGN:
 
         return d2
 
-    class SigningKeyBase(object):
+    @util.public
+    class CryptosignKey(object):
+        """
+        A cryptosign private key for signing, and hence usable for authentication or a
+        public key usable for verification (but can't be used for signing).
+        """
 
-        def __init__(self, signer, can_sign: bool) -> None:
+        def __init__(self, key, can_sign: bool, security_module: Optional[ISecurityModule] = None,
+                     key_no: Optional[int] = None, comment: Optional[str] = None) -> None:
+            if not (isinstance(key, signing.VerifyKey) or isinstance(key, signing.SigningKey)):
+                raise Exception("invalid type {} for key".format(type(key)))
+
+            assert (can_sign and isinstance(key, signing.SigningKey)) or (not can_sign and isinstance(key, signing.VerifyKey))
+            self._key = key
             self._can_sign = can_sign
-            self._key = signer
+            self._security_module = security_module
+            self._key_no = key_no
+            self._comment = comment
 
-        @util.public
-        def can_sign(self):
+        @property
+        def security_module(self) -> Optional['ISecurityModule']:
             """
-            Check if the key can be used to sign.
+            Implements :meth:`autobahn.wamp.interfaces.IKey.security_module`.
+            """
+            return self._security_module
 
-            :returns: `True`, iff the key can sign.
-            :rtype: bool
+        @property
+        def key_no(self) -> Optional[int]:
+            """
+            Implements :meth:`autobahn.wamp.interfaces.IKey.key_no`.
+            """
+            return self._key_no
+
+        @property
+        def comment(self) -> Optional[str]:
+            """
+            Implements :meth:`autobahn.wamp.interfaces.IKey.comment`.
+            """
+            return self._comment
+
+        @property
+        def key_type(self) -> str:
+            """
+            Implements :meth:`autobahn.wamp.interfaces.IKey.key_type`.
+            """
+            return 'cryptosign'
+
+        @property
+        def can_sign(self) -> bool:
+            """
+            Implements :meth:`autobahn.wamp.interfaces.IKey.can_sign`.
             """
             return self._can_sign
 
-        @util.public
-        def sign_challenge(self, session: ISession, challenge: Challenge, channel_id_type: Optional[str] = None) -> bytes:
-            """
-            Sign WAMP-cryptosign challenge.
-
-            :param session: The authenticating WAMP session.
-            :param challenge: The WAMP-cryptosign challenge object for which a signature should be computed.
-            :returns: A Deferred/Future that resolves to the computed signature.
-            """
-            # get the TLS channel ID of the underlying TLS connection
-            channel_id_type = 'tls-unique'
-            if channel_id_type in session._transport.transport_details.channel_id:
-                channel_id = session._transport.transport_details.channel_id.get(channel_id_type, None)
-            else:
-                channel_id_type = None
-                channel_id = None
-
-            data = format_challenge(challenge, channel_id, channel_id_type)
-
-            return sign_challenge(data, self.sign)
-
         @util.public
         def sign(self, data: bytes) -> bytes:
             """
-            Sign some data.
-
-            :param data: The data to be signed.
-            :type data: bytes
-
-            :returns: The signature.
-            :rtype: bytes
+            Implements :meth:`autobahn.wamp.interfaces.IKey.sign`.
             """
             if not self._can_sign:
                 raise Exception("a signing key required to sign")
@@ -484,8 +500,7 @@ if HAS_CRYPTOSIGN:
             if type(data) != bytes:
                 raise Exception("data to be signed must be binary")
 
-            # sig is a nacl.signing.SignedMessage
-            sig = self._key.sign(data)
+            sig: SignedMessage = self._key.sign(data)
 
             # we only return the actual signature! if we return "sig",
             # it gets coerced into the concatenation of message + signature
@@ -493,44 +508,18 @@ if HAS_CRYPTOSIGN:
             # the signature
             return txaio.create_future_success(sig.signature)
 
-    @util.public
-    class SigningKey(SigningKeyBase):
-        """
-        A cryptosign private key for signing, and hence usable for authentication or a
-        public key usable for verification (but can't be used for signing).
-        """
-
-        def __init__(self, key, comment=None):
-            """
-
-            :param key: A Ed25519 private signing key or a Ed25519 public verification key.
-            :type key: instance of nacl.signing.VerifyKey or instance of nacl.signing.SigningKey
-            """
-            if not (isinstance(key, signing.VerifyKey) or isinstance(key, signing.SigningKey)):
-                raise Exception("invalid type {} for key".format(type(key)))
-
-            if not (comment is None or type(comment) == str):
-                raise Exception("invalid type {} for comment".format(type(comment)))
-
-            self._key = key
-            self._comment = comment
-            can_sign = isinstance(key, signing.SigningKey)
-
-            super().__init__(key, can_sign)
-
-        def __str__(self):
-            comment = '"{}"'.format(self.comment()) if self.comment() else None
-            return 'Key(can_sign={}, comment={}, public_key={})'.format(self.can_sign(), comment, self.public_key())
-
         @util.public
-        def comment(self):
+        def sign_challenge(self, challenge: Challenge, channel_id: Optional[bytes] = None,
+                           channel_id_type: Optional[str] = None) -> bytes:
             """
-            Get the key comment (if any).
+            Implements :meth:`autobahn.wamp.interfaces.ICryptosignKey.sign_challenge`.
+            """
+            assert challenge.method in ['cryptosign', 'cryptosign-proxy'], \
+                'unexpected cryptosign challenge with method "{}"'.format(challenge.method)
 
-            :returns: The comment (if any) from the key.
-            :rtype: str or None
-            """
-            return self._comment
+            data = _format_challenge(challenge, channel_id, channel_id_type)
+
+            return _sign_challenge(data, self.sign)
 
         @util.public
         def public_key(self, binary: bool = False) -> Union[str, bytes]:
@@ -552,22 +541,22 @@ if HAS_CRYPTOSIGN:
 
         @util.public
         @classmethod
-        def from_key_bytes(cls, keydata: str, comment: Optional[str] = None) -> 'SigningKey':
+        def from_bytes(cls, key_data: bytes, comment: Optional[str] = None) -> 'CryptosignKey':
             if not (comment is None or type(comment) == str):
                 raise ValueError("invalid type {} for comment".format(type(comment)))
 
-            if type(keydata) != bytes:
-                raise ValueError("invalid key type {} (expected binary)".format(type(keydata)))
+            if type(key_data) != bytes:
+                raise ValueError("invalid key type {} (expected binary)".format(type(key_data)))
 
-            if len(keydata) != 32:
-                raise ValueError("invalid key length {} (expected 32)".format(len(keydata)))
+            if len(key_data) != 32:
+                raise ValueError("invalid key length {} (expected 32)".format(len(key_data)))
 
-            key = signing.SigningKey(keydata)
-            return cls(key, comment)
+            key = signing.SigningKey(key_data)
+            return cls(key=key, can_sign=True, comment=comment)
 
         @util.public
         @classmethod
-        def from_raw_key(cls, filename: str, comment: Optional[str] = None) -> 'SigningKey':
+        def from_file(cls, filename: str, comment: Optional[str] = None) -> 'CryptosignKey':
             """
             Load an Ed25519 (private) signing key (actually, the seed for the key) from a raw file of 32 bytes length.
             This can be any random byte sequence, such as generated from Python code like
@@ -579,10 +568,8 @@ if HAS_CRYPTOSIGN:
                 dd if=/dev/urandom of=client02.key bs=1 count=32
 
             :param filename: Filename of the key.
-            :type filename: str
             :param comment: Comment for key (optional).
-            :type comment: str or None
-            """
+           """
             if not (comment is None or type(comment) == str):
                 raise Exception("invalid type {} for comment".format(type(comment)))
 
@@ -590,13 +577,13 @@ if HAS_CRYPTOSIGN:
                 raise Exception("invalid type {} for filename".format(filename))
 
             with open(filename, 'rb') as f:
-                keydata = f.read()
+                key_data = f.read()
 
-            return cls.from_key_bytes(keydata, comment=comment)
+            return cls.from_bytes(key_data, comment=comment)
 
         @util.public
         @classmethod
-        def from_ssh_key(cls, filename: str) -> 'SigningKey':
+        def from_ssh_file(cls, filename: str) -> 'CryptosignKey':
             """
             Load an Ed25519 key from a SSH key file. The key file can be a (private) signing
             key (from a SSH private key file) or a (public) verification key (from a SSH
@@ -604,27 +591,379 @@ if HAS_CRYPTOSIGN:
             """
 
             with open(filename, 'rb') as f:
-                keydata = f.read().decode('utf-8').strip()
-            return cls.from_ssh_data(keydata)
+                key_data = f.read().decode('utf-8').strip()
+            return cls.from_ssh_bytes(key_data)
 
         @util.public
         @classmethod
-        def from_ssh_data(cls, keydata: str) -> 'SigningKey':
+        def from_ssh_bytes(cls, key_data: str) -> 'CryptosignKey':
             """
             Load an Ed25519 key from SSH key file. The key file can be a (private) signing
             key (from a SSH private key file) or a (public) verification key (from a SSH
             public key file). A private key file must be passphrase-less.
             """
             SSH_BEGIN = '-----BEGIN OPENSSH PRIVATE KEY-----'
-            if keydata.startswith(SSH_BEGIN):
+            if key_data.startswith(SSH_BEGIN):
                 # OpenSSH private key
-                keydata, comment = _read_ssh_ed25519_privkey(keydata)
-                key = signing.SigningKey(keydata, encoder=encoding.RawEncoder)
+                key_data, comment = _read_ssh_ed25519_privkey(key_data)
+                key = signing.SigningKey(key_data, encoder=encoding.RawEncoder)
+                can_sign = True
             else:
                 # OpenSSH public key
-                keydata, comment = _read_ssh_ed25519_pubkey(keydata)
-                key = signing.VerifyKey(keydata)
+                key_data, comment = _read_ssh_ed25519_pubkey(key_data)
+                key = signing.VerifyKey(key_data)
+                can_sign = False
 
-            return cls(key, comment)
+            return cls(key=key, can_sign=can_sign, comment=comment)
 
-    IEd25519Key.register(SigningKey)
+        @classmethod
+        def from_seedphrase(cls, seedphrase: str, index: int = 0) -> 'CryptosignKey':
+            """
+            Create a private key from the given BIP-39 mnemonic seed phrase and index,
+            which can be used to sign and create signatures.
+
+            :param seedphrase: The BIP-39 seedphrase ("Mnemonic") from which to derive the account.
+            :param index: The account index in account hierarchy defined by the seedphrase.
+            :return: New instance of :class:`EthereumKey`
+            """
+            try:
+                from autobahn.xbr._mnemonic import mnemonic_to_private_key
+            except ImportError as e:
+                raise RuntimeError('package autobahn[xbr] not installed ("{}")'.format(e))
+
+            # BIP44 path for WAMP
+            # https://github.com/wamp-proto/wamp-proto/issues/401
+            # https://github.com/satoshilabs/slips/pull/1322
+            derivation_path = "m/44'/655'/0'/0/{}".format(index)
+
+            key_raw = mnemonic_to_private_key(seedphrase, derivation_path)
+            assert type(key_raw) == bytes
+            assert len(key_raw) == 32
+
+            # create WAMP-Cryptosign key object from raw bytes
+            key = cls.from_bytes(key_raw)
+
+            return key
+
+    ICryptosignKey.register(CryptosignKey)
+
+    class CryptosignAuthextra(object):
+        """
+        WAMP-Cryptosign authextra object.
+        """
+        __slots__ = [
+            '_pubkey',
+            '_trustroot',
+            '_challenge',
+            '_channel_binding',
+            '_channel_id',
+            '_realm',
+            '_chain_id',
+            '_block_no',
+            '_delegate',
+            '_seeder',
+            '_bandwidth',
+            '_signature',
+        ]
+
+        def __init__(self,
+                     pubkey: Optional[bytes] = None,
+                     challenge: Optional[bytes] = None,
+                     channel_binding: Optional[str] = None,
+                     channel_id: Optional[bytes] = None,
+
+                     # domain address, certificates are verified against owner of the domain
+                     trustroot: Optional[bytes] = None,
+
+                     # FIXME: add delegate address
+                     # FIXME: add certificates
+                     # FIXME: remove reservation
+                     realm: Optional[bytes] = None,
+                     chain_id: Optional[int] = None,
+                     block_no: Optional[int] = None,
+                     delegate: Optional[bytes] = None,
+                     seeder: Optional[bytes] = None,
+                     bandwidth: Optional[int] = None,
+
+                     signature: Optional[bytes] = None,
+                     ):
+            if pubkey:
+                assert len(pubkey) == 32
+            if trustroot:
+                assert len(trustroot) == 20
+            if challenge:
+                assert len(challenge) == 32
+            if channel_binding:
+                assert channel_binding in ['tls-unique']
+            if channel_id:
+                assert len(channel_id) == 32
+            if realm:
+                assert len(realm) == 20
+            if delegate:
+                assert len(delegate) == 20
+            if seeder:
+                assert len(seeder) == 20
+            if signature:
+                assert len(signature) == 65
+            self._pubkey = pubkey
+            self._trustroot = trustroot
+            self._challenge = challenge
+            self._channel_binding = channel_binding
+            self._channel_id = channel_id
+            self._realm = realm
+            self._chain_id = chain_id
+            self._block_no = block_no
+            self._delegate = delegate
+            self._seeder = seeder
+            self._bandwidth = bandwidth
+            self._signature = signature
+
+        @property
+        def pubkey(self) -> Optional[bytes]:
+            return self._pubkey
+
+        @pubkey.setter
+        def pubkey(self, value: Optional[bytes]):
+            assert value is None or len(value) == 20
+            self._pubkey = value
+
+        @property
+        def trustroot(self) -> Optional[bytes]:
+            return self._trustroot
+
+        @trustroot.setter
+        def trustroot(self, value: Optional[bytes]):
+            assert value is None or len(value) == 20
+            self._trustroot = value
+
+        @property
+        def challenge(self) -> Optional[bytes]:
+            return self._challenge
+
+        @challenge.setter
+        def challenge(self, value: Optional[bytes]):
+            assert value is None or len(value) == 32
+            self._challenge = value
+
+        @property
+        def channel_binding(self) -> Optional[str]:
+            return self._channel_binding
+
+        @channel_binding.setter
+        def channel_binding(self, value: Optional[str]):
+            assert value is None or value in ['tls-unique']
+            self._channel_binding = value
+
+        @property
+        def channel_id(self) -> Optional[bytes]:
+            return self._channel_id
+
+        @channel_id.setter
+        def channel_id(self, value: Optional[bytes]):
+            assert value is None or len(value) == 32
+            self._channel_id = value
+
+        @property
+        def realm(self) -> Optional[bytes]:
+            return self._realm
+
+        @realm.setter
+        def realm(self, value: Optional[bytes]):
+            assert value is None or len(value) == 20
+            self._realm = value
+
+        @property
+        def chain_id(self) -> Optional[int]:
+            return self._chain_id
+
+        @chain_id.setter
+        def chain_id(self, value: Optional[int]):
+            assert value is None or value > 0
+            self._chain_id = value
+
+        @property
+        def block_no(self) -> Optional[int]:
+            return self._block_no
+
+        @block_no.setter
+        def block_no(self, value: Optional[int]):
+            assert value is None or value > 0
+            self._block_no = value
+
+        @property
+        def delegate(self) -> Optional[bytes]:
+            return self._delegate
+
+        @delegate.setter
+        def delegate(self, value: Optional[bytes]):
+            assert value is None or len(value) == 20
+            self._delegate = value
+
+        @property
+        def seeder(self) -> Optional[bytes]:
+            return self._seeder
+
+        @seeder.setter
+        def seeder(self, value: Optional[bytes]):
+            assert value is None or len(value) == 20
+            self._seeder = value
+
+        @property
+        def bandwidth(self) -> Optional[int]:
+            return self._bandwidth
+
+        @bandwidth.setter
+        def bandwidth(self, value: Optional[int]):
+            assert value is None or value > 0
+            self._bandwidth = value
+
+        @property
+        def signature(self) -> Optional[bytes]:
+            return self._signature
+
+        @signature.setter
+        def signature(self, value: Optional[bytes]):
+            assert value is None or len(value) == 65
+            self._signature = value
+
+        @staticmethod
+        def parse(data: Dict[str, Any]) -> 'CryptosignAuthextra':
+            obj = CryptosignAuthextra()
+
+            pubkey = data.get('pubkey', None)
+            if pubkey is not None:
+                if type(pubkey) != str:
+                    raise ValueError('invalid type {} for pubkey'.format(type(pubkey)))
+                if len(pubkey) != 32 * 2:
+                    raise ValueError('invalid length {} of pubkey'.format(len(pubkey)))
+                obj._pubkey = a2b_hex(pubkey)
+
+            challenge = data.get('challenge', None)
+            if challenge is not None:
+                if type(challenge) != str:
+                    raise ValueError('invalid type {} for challenge'.format(type(challenge)))
+                if len(challenge) != 32 * 2:
+                    raise ValueError('invalid length {} of challenge'.format(len(challenge)))
+                obj._challenge = a2b_hex(challenge)
+
+            channel_binding = data.get('channel_binding', None)
+            if channel_binding is not None:
+                if type(channel_binding) != str:
+                    raise ValueError('invalid type {} for channel_binding'.format(type(channel_binding)))
+                if channel_binding not in ['tls-unique']:
+                    raise ValueError('invalid value "{}" for channel_binding'.format(channel_binding))
+                obj._channel_binding = channel_binding
+
+            channel_id = data.get('channel_id', None)
+            if channel_id is not None:
+                if type(channel_id) != str:
+                    raise ValueError('invalid type {} for channel_id'.format(type(channel_id)))
+                if len(channel_id) != 32 * 2:
+                    raise ValueError('invalid length {} of channel_id'.format(len(channel_id)))
+                obj._channel_id = a2b_hex(channel_id)
+
+            trustroot = data.get('trustroot', None)
+            if trustroot is not None:
+                if type(trustroot) != str:
+                    raise ValueError('invalid type {} for trustroot - expected a string'.format(type(trustroot)))
+                if not _URI_PAT_REALM_NAME_ETH.match(trustroot):
+                    raise ValueError('invalid value "{}" for trustroot - expected an Ethereum address'.format(type(trustroot)))
+                obj._trustroot = a2b_hex(trustroot[2:])
+
+            reservation = data.get('reservation', None)
+            if reservation is not None:
+                if type(reservation) != dict:
+                    raise ValueError('invalid type {} for reservation'.format(type(reservation)))
+
+                chain_id = reservation.get('chain_id', None)
+                if chain_id is not None:
+                    if type(chain_id) != int:
+                        raise ValueError('invalid type {} for reservation.chain_id - expected an integer'.format(type(chain_id)))
+                    obj._chain_id = chain_id
+
+                block_no = reservation.get('block_no', None)
+                if block_no is not None:
+                    if type(block_no) != int:
+                        raise ValueError('invalid type {} for reservation.block_no - expected an integer'.format(type(block_no)))
+                    obj._block_no = block_no
+
+                realm = reservation.get('realm', None)
+                if realm is not None:
+                    if type(realm) != str:
+                        raise ValueError('invalid type {} for reservation.realm - expected a string'.format(type(realm)))
+                    if not _URI_PAT_REALM_NAME_ETH.match(realm):
+                        raise ValueError('invalid value "{}" for reservation.realm - expected an Ethereum address'.format(type(realm)))
+                    obj._realm = a2b_hex(realm[2:])
+
+                delegate = reservation.get('delegate', None)
+                if delegate is not None:
+                    if type(delegate) != str:
+                        raise ValueError('invalid type {} for reservation.delegate - expected a string'.format(type(delegate)))
+                    if not _URI_PAT_REALM_NAME_ETH.match(delegate):
+                        raise ValueError('invalid value "{}" for reservation.delegate - expected an Ethereum address'.format(type(delegate)))
+                    obj._delegate = a2b_hex(delegate[2:])
+
+                seeder = reservation.get('seeder', None)
+                if seeder is not None:
+                    if type(seeder) != str:
+                        raise ValueError('invalid type {} for reservation.seeder - expected a string'.format(type(seeder)))
+                    if not _URI_PAT_REALM_NAME_ETH.match(seeder):
+                        raise ValueError('invalid value "{}" for reservation.seeder - expected an Ethereum address'.format(type(seeder)))
+                    obj._seeder = a2b_hex(seeder[2:])
+
+                bandwidth = reservation.get('bandwidth', None)
+                if bandwidth is not None:
+                    if type(bandwidth) != int:
+                        raise ValueError('invalid type {} for reservation.bandwidth - expected an integer'.format(type(bandwidth)))
+                    obj._bandwidth = bandwidth
+
+            signature = data.get('signature', None)
+            if signature is not None:
+                if type(signature) != str:
+                    raise ValueError('invalid type {} for signature'.format(type(signature)))
+                if len(signature) != 65 * 2:
+                    raise ValueError('invalid length {} of signature'.format(len(signature)))
+                obj._signature = a2b_hex(signature)
+
+            return obj
+
+        def marshal(self) -> Dict[str, Any]:
+            res = {}
+
+            # FIXME: marshal check-summed eth addresses
+
+            if self._pubkey is not None:
+                res['pubkey'] = b2a_hex(self._pubkey).decode()
+
+            if self._challenge is not None:
+                res['challenge'] = b2a_hex(self._challenge).decode()
+            if self._channel_binding is not None:
+                res['channel_binding'] = self._channel_binding
+            if self._channel_id is not None:
+                res['channel_id'] = b2a_hex(self._channel_id).decode()
+
+            if self._trustroot is not None:
+                res['trustroot'] = '0x' + b2a_hex(self._trustroot).decode()
+
+            reservation = {}
+            if self._chain_id is not None:
+                reservation['chain_id'] = self._chain_id
+            if self._block_no is not None:
+                reservation['block_no'] = self._block_no
+            if self._realm is not None:
+                reservation['realm'] = '0x' + b2a_hex(self._realm).decode()
+            if self._delegate is not None:
+                reservation['delegate'] = '0x' + b2a_hex(self._delegate).decode()
+            if self._seeder is not None:
+                reservation['seeder'] = '0x' + b2a_hex(self._seeder).decode()
+            if self._bandwidth is not None:
+                reservation['bandwidth'] = self._bandwidth
+            if reservation:
+                res['reservation'] = reservation
+
+            if self._signature is not None:
+                res['signature'] = b2a_hex(self._signature).decode()
+
+            return res
+
+    __all__.extend(['CryptosignKey', 'format_challenge', 'sign_challenge', 'CryptosignAuthextra'])
